<?php
namespace nulib\ext\spout;

use nulib\cl;
use nulib\file\tab\AbstractBuilder;
use nulib\file\tab\TAbstractBuilder;
use nulib\os\path;
use nulib\php\func;
use nulib\php\nur_func;
use nulib\php\time\Date;
use nulib\php\time\DateTime;
use nulib\ref\ext\spout\ref_builder;
use nulib\ref\ext\spout\ref_builder_ods;
use nulib\ref\ext\spout\ref_builder_xlsx;
use nulib\str;
use nulib\web\http;
use OpenSpout\Common\Entity\Style\Border;
use OpenSpout\Common\Entity\Style\BorderPart;
use OpenSpout\Common\Entity\Cell\DateTimeCell;
use OpenSpout\Common\Entity\Cell\EmptyCell;
use OpenSpout\Common\Entity\Cell\NumericCell;
use OpenSpout\Common\Entity\Cell\StringCell;
use OpenSpout\Common\Entity\Row;
use OpenSpout\Common\Entity\Style\Color;
use OpenSpout\Common\Entity\Style\Style;
use OpenSpout\Writer\AbstractWriterMultiSheets;
use OpenSpout\Writer\Common\AbstractOptions;
use OpenSpout\Writer\ODS\Options as ODSOptions;
use OpenSpout\Writer\ODS\Writer as ODSWriter;
use OpenSpout\Writer\XLSX\Entity\SheetView;
use OpenSpout\Writer\XLSX\Options as XLSXOptions;
use OpenSpout\Writer\XLSX\Writer as XLSXWriter;

class SpoutBuilder extends AbstractBuilder {
  use TAbstractBuilder;

  protected static function apply_params($object, ?array $params, array $refParams) {
    foreach (array_keys($refParams) as $method) {
      if (!str::starts_with("->", $method)) continue;
      $func = func::with([$object, $method]);
      if (($args = $params[$method] ?? null) !== null) {
        $func->invoke(cl::with($args));
      }
      if (($argss = $params["$method*"] ?? null) !== null) {
        foreach ($argss as $args) {
          $func->invoke(cl::with($args));
        }
      }
    }
    return $object;
  }

  protected static function add_border_part(?array &$parts, string $name, ?array $params): void {
    if ($params === null) return;
    if (($color = $params["color"] ?? null) !== null) {
      $color = cl::get(ref_builder::COLORS, $color, $color);
    } else {
      $color = Color::BLACK;
    }
    if (($width = $params["width"] ?? null) === null) $width = Border::WIDTH_MEDIUM;
    if (($style = $params["style"] ?? null) === null) $style = Border::STYLE_SOLID;
    $parts[] = new BorderPart($name, $color, $width, $style);
  }

  protected static function set_defaults(?array &$params, string $key, array $defaults): void {
    if ($params !== null && array_key_exists($key, $params)) {
      if ($params[$key] === false) $params[$key] = null;
      else $params[$key] ??= $defaults;
    } else {
      $params[$key] ??= $defaults;
    }
  }

  protected static function ensure_style(&$style): ?Style {
    if ($style === null) return null;
    if ($style instanceof Style) return $style;
    $cell = $style;
    $style = new Style();
    $font = $cell["font"] ?? null;
    if ($font["bold"] ?? null) $style->setFontBold();
    if ($font["italic"] ?? null) $style->setFontItalic();
    if ($font["underline"] ?? null) $style->setFontUnderline();
    if ($font["strikethrough"] ?? null) $style->setFontStrikethrough();
    if (($name = $font["name"] ?? null) !== null) $style->setFontName($name);
    if (($size = $font["size"] ?? null) !== null) $style->setFontSize($size);
    if (($color = $font["color"] ?? null) !== null) {
      $style->setFontColor(cl::get(ref_builder::COLORS, $color, $color));
    }
    if (($color = $cell["bg_color"] ?? null) !== null) {
      $style->setBackgroundColor(cl::get(ref_builder::COLORS, $color, $color));
    }
    if (($align = $cell["align"] ?? null) !== null) $style->setCellAlignment($align);
    if (($align = $cell["valign"] ?? null) !== null) $style->setCellVerticalAlignment($align);
    if (($wrap = $cell["wrap"] ?? null) !== null) $style->setShouldWrapText($wrap);
    if (($rotation = $cell["rotation"] ?? null) !== null) $style->setTextRotation($rotation);
    if (($format = $cell["format"] ?? null) !== null) $style->setFormat($format);
    if (($border = $cell["border"] ?? null) !== null) {
      if (is_string($border)) {
        $parts = explode(" ", $border);
        $border = [];
        $styleAll = null;
        $widthAll = null;
        $colorAll = null;
        foreach ($parts as $part) {
          if ($part === "all") {
            $border["left"] = [];
            $border["top"] = [];
            $border["right"] = [];
            $border["bottom"] = [];
          } elseif (preg_match('/^(left|top|right|bottom)$/', $part)) {
            $border[$part] = [];
          } elseif (preg_match('/^(none|solid|dashed|dotted|double)$/', $part)) {
            $styleAll = $part;
          } elseif (preg_match('/^(thin|medium|thick)$/', $part)) {
            $widthAll = $part;
          } else {
            $colorAll = $part;
          }
        }
        foreach ($border as &$part) {
          if ($styleAll !== null) $part["style"] = $styleAll;
          if ($widthAll !== null) $part["width"] = $widthAll;
          if ($colorAll !== null) $part["color"] = $colorAll;
        }; unset($part);
      }
      $parts = null;
      self::add_border_part($parts, "top", $border["top"] ?? null);
      self::add_border_part($parts, "right", $border["right"] ?? null);
      self::add_border_part($parts, "bottom", $border["bottom"] ?? null);
      self::add_border_part($parts, "left", $border["left"] ?? null);
      if ($parts !== null) $style->setBorder(new Border(...$parts));
    }
    if (($autofit = $cell["autofit"] ?? null) !== null) $style->setShouldShrinkToFit($autofit);
    return $style;
  }

  const DATE_FORMAT = "dd/mm/yyyy";

  const DATETIME_FORMAT = "dd/mm/yyyy hh:mm:ss";

  /** @var bool faut-il choisir le type numérique pour une chaine numérique? */
  const TYPE_NUMERIC = true;

  /** @var bool faut-il choisir le type date pour une chaine au bon format? */
  const TYPE_DATE = true;

  /** @var array configuration du writer */
  const SPOUT_PARAMS = null;

  /** @var array configuration de la première feuille */
  const SHEET_PARAMS = null;

  /** @var string nom de la première feuille */
  const SHEET_NAME = null;

  /** @var array configuration de la vue de la première feuille */
  const SHEET_VIEW_PARAMS = null;

  function __construct(?string $output, ?array $params=null) {
    parent::__construct($output, $params);
    $ssType = $params["ss_type"] ?? null;
    if ($ssType === null) {
      switch (path::ext($this->output)) {
      case ".ods":
        $ssType = self::SS_TYPE_ODS;
        break;
      case ".xlsx":
      default:
        $ssType = self::SS_TYPE_XLSX;
        break;
      }
    }

    $spoutParams = $params["spout"] ?? static::SPOUT_PARAMS;
    $spoutParams["default_column_width"] ??= 10.5;
    self::ensure_style($spoutParams["default_row_style"]);
    switch ($ssType) {
    case "ods":
    case self::SS_TYPE_ODS:
      $ssType = self::SS_TYPE_ODS;
      $ssOptions = new ODSOptions();
      $ssWriterClass = ODSWriter::class;
      $refParams = ref_builder_ods::PARAMS_SPOUT;
      break;
    case "xlsx":
    case self::SS_TYPE_XLSX:
    default:
        $ssType = self::SS_TYPE_XLSX;
        $ssOptions = new XLSXOptions();
        $ssWriterClass = XLSXWriter::class;
        $refParams = ref_builder_xlsx::PARAMS_SPOUT;
        break;
    }
    $defaultColumnWidth = $spoutParams["default_column_width"] ?? null;
    if ($defaultColumnWidth !== null) $ssOptions->DEFAULT_COLUMN_WIDTH = $defaultColumnWidth;
    $defaultRowHeight = $spoutParams["default_row_height"] ?? null;
    if ($defaultRowHeight !== null) $ssOptions->DEFAULT_ROW_HEIGHT = $defaultRowHeight;
    $defaultRowStyle = $spoutParams["default_row_style"] ?? null;
    if ($defaultRowStyle !== null) $ssOptions->DEFAULT_ROW_STYLE = $defaultRowStyle;
    self::apply_params($ssOptions, $spoutParams, $refParams);
    $ssWriter = new $ssWriterClass($ssOptions);
    $ssWriter->writeToStream($this->getResource());

    $this->ssType = $ssType;
    $this->ssOptions = $ssOptions;
    $this->ssWriter = $ssWriter;
    $this->spoutParams = $spoutParams;
    $this->typeNumeric = boolval($params["type_numeric"] ?? static::TYPE_NUMERIC);
    $this->typeDate = boolval($params["type_date"] ?? static::TYPE_DATE);

    $sheetParams = $params["sheet"] ?? static::SHEET_PARAMS;
    $sheetName = $params["sheet_name"] ?? static::SHEET_NAME;
    if ($sheetName !== null) $sheetParams["->setName"] = $sheetName;
    $sheetViewParams = $params["sheet_view"] ?? static::SHEET_VIEW_PARAMS;
    if ($sheetViewParams !== null) $sheetParams["view"] = $sheetViewParams;
    $this->firstSheet = true;
    $this->sheetParams = null;
    $this->setSheet(null, $sheetParams);
  }

  const SS_TYPE_ODS = 1, SS_TYPE_XLSX = 2;

  /** @var int type de fichier généré */
  protected int $ssType;

  protected AbstractOptions $ssOptions;

  protected AbstractWriterMultiSheets $ssWriter;

  protected ?array $spoutParams;

  protected bool $typeNumeric;

  protected bool $typeDate;

  function setParams(?array $params): self {
    if ($params !== null) {
      if (array_key_exists("type_numeric", $params)) {
        $this->typeNumeric = boolval($params["type_numeric"] ?? static::TYPE_NUMERIC);
      }
      if (array_key_exists("type_date", $params)) {
        $this->typeDate = boolval($params["type_date"] ?? static::TYPE_DATE);
      }
    }
    return $this;
  }

  protected bool $firstSheet;

  protected ?array $sheetParams;

  const STYLE_ROW = 0, STYLE_HEADER = 1;

  protected int $rowStyle;

  protected int $currentRow;

  protected ?bool $differentOddEven = null;

  protected int $oddEvenIndex = 0;

  function setDifferentOddEven(bool $differentOddEven, ?bool $startWithOdd=null): self {
    $this->differentOddEven = $differentOddEven;
    if ($differentOddEven && $startWithOdd !== null) $this->oddEvenIndex = $startWithOdd? 1: 0;
    return $this;
  }

  /**
   * @param string|int|null $sheetName
   */
  function setSheet($sheetName, ?array $sheetParams=null): self {
    $sheet = $params["sheet"] ?? null;
    $sheetName = $sheetName ?? $sheetParams["sheet_name"] ?? null;
    $sheetViewParams = $sheetParams["sheet_view"] ?? null;

    if ($sheet !== null) $sheetParams = $sheet;
    if ($sheetName !== null) $sheetParams["->setName"] = $sheetName;
    if ($sheetViewParams !== null) $sheetParams["view"] = $sheetViewParams;

    $writer = $this->ssWriter;
    if ($this->firstSheet) {
      $this->firstSheet = false;
      $sheet = $writer->getCurrentSheet();
    } else {
      $sheet = $writer->addNewSheetAndMakeItCurrent();
      $this->wroteHeaders = false;
      $this->built = false;
    }
    $this->rowStyle = self::STYLE_ROW;
    $this->currentRow = 1;

    switch ($this->ssType) {
    case self::SS_TYPE_ODS:
      # appliquer les paramètres de la feuille
      $this->apply_params($sheet, $sheetParams, ref_builder_ods::PARAMS_SHEET);
      break;
    case self::SS_TYPE_XLSX:
      # appliquer les paramètres de la feuille
      $this->apply_params($sheet, $sheetParams, ref_builder_xlsx::PARAMS_SHEET);
      # appliquer les paramètres de la vue de la feuille
      $sheetViewParams =& $sheetParams["view"];
      $sheetViewParams["->setFreezeRow"] ??= 2;
      $sheet->setSheetView(self::apply_params(new SheetView(), $sheetViewParams, ref_builder_xlsx::PARAMS_SHEET_VIEW));
      break;
    }
    self::set_defaults($sheetParams, "header_style", [
      "font" => ["bold" => true],
      "bg_color" => "gray",
    ]);
    self::set_defaults($sheetParams, "odd_style", [
      "wrap" => false,
    ]);
    self::set_defaults($sheetParams, "even_style", [
      "bg_color" => "light_gray",
      "wrap" => false,
    ]);
    $this->ensure_style($sheetParams["header_style"]);
    $this->ensure_style($sheetParams["odd_style"]);
    $this->ensure_style($sheetParams["even_style"]);
    $this->sheetParams = $sheetParams;

    if ($sheetParams !== null) {
      if (array_key_exists("schema", $sheetParams)) {
        $this->schema = $sheetParams["schema"] ?? null;
      }
      if (array_key_exists("headers", $sheetParams)) {
        $this->headers = $sheetParams["headers"] ?? null;
      }
      if (array_key_exists("rows", $sheetParams)) {
        $rows = $sheetParams["rows"] ?? null;
        if (is_callable($rows)) $rows = $rows();
        $this->rows = $rows;
      }
      if (array_key_exists("cook_func", $sheetParams)) {
        $cookFunc = $sheetParams["cook_func"] ?? null;
        $cookCtx = $cookArgs = null;
        if ($cookFunc !== null) {
          nur_func::ensure_func($cookFunc, $this, $cookArgs);
          $cookCtx = nur_func::_prepare($cookFunc);
        }
        $this->cookCtx = $cookCtx;
        $this->cookArgs = $cookArgs;
      }
    }
    return $this;
  }

  /**
   * les colonnes sont indexées sur 0 (e.g A = 0, B = 1, etc.)
   * Les lignes sont indexées sur 1
   */
  function mergeCells(int $topLeftCol, int $topLeftRow, int $bottomRightCol, int $bottomRightRow): void {
    $sheet = $this->ssWriter->getCurrentSheet();
    $this->ssOptions->mergeCells($topLeftCol, $topLeftRow, $bottomRightCol, $bottomRightRow, $sheet->getIndex());
  }

  protected function isNumeric($value): bool {
    if ($this->typeNumeric && is_numeric($value)) return true;
    if (!is_string($value) && is_numeric($value)) return true;
    return false;
  }

  protected function isDate(&$value, &$style): bool {
    if ($value instanceof Date) {
      $style ??= new Style();
      $style->setFormat(self::DATE_FORMAT);
      return true;
    } elseif ($value instanceof DateTime) {
      $style ??= new Style();
      $style->setFormat(self::DATETIME_FORMAT);
      return true;
    }
    if (!is_string($value) || !$this->typeDate) return false;
    if (DateTime::isa_datetime($value, true)) {
      $value = new DateTime($value);
      $style ??= new Style();
      $style->setFormat(self::DATETIME_FORMAT);
      return true;
    }
    if (DateTime::isa_date($value, true)) {
      $value = new Date($value);
      $style ??= new Style();
      $style->setFormat(self::DATE_FORMAT);
      return true;
    }
    return false;
  }

  function _write(array $row, ?array $colStyles=null, ?array $rowStyle=null): void {
    $rowParams = null;
    if ($rowStyle !== null) {
      # séparer rowParams (pour configurer l'instance de $row) et $rowStyle
      # (pour appliquer un style sur la ligne)
      foreach (array_keys(ref_builder::ROW_PARAMS) as $method) {
        $value = $rowStyle[$method] ?? null;
        unset($rowStyle[$method]);
        if ($value !== null) $rowParams[$method] = $value;
      }
      if ($rowStyle === []) $rowStyle = null;
    }
    $sheetParams = $this->sheetParams;
    $headerStyle = $sheetParams["header_style"] ?? null;
    $oddStyle = $sheetParams["odd_style"] ?? null;
    $evenStyle = $sheetParams["even_style"] ?? null;
    $differentOddEven = $this->differentOddEven;
    $differentOddEven ??= $sheetParams["different_odd_even"] ?? true;

    $cells = [];
    foreach ($row as $key => $col) {
      $style = $colStyles[$key] ?? null;
      self::ensure_style($style);
      if ($col === null || $col === "") {
        $cell = new EmptyCell(null, $style);
      } elseif ($this->isNumeric($col)) {
        $cell = new NumericCell($col, $style);
      } elseif ($this->isDate($col, $style)) {
        $cell = new DateTimeCell($col, $style);
      } else {
        $cell = new StringCell($col, $style);
      }
      $cells[] = $cell;
    }

    if ($this->rowStyle === self::STYLE_HEADER) {
      $rowStyle ??= $headerStyle;
    } elseif ($differentOddEven && $this->oddEvenIndex % 2 == 0) {
      $rowStyle ??= $evenStyle;
    }
    $rowStyle ??= $oddStyle;
    self::ensure_style($rowStyle);
    $row = new Row($cells, $rowStyle);
    self::apply_params($row, $rowParams, ref_builder::ROW_PARAMS);

    $mergeCells = $rowParams["merge_cells"] ?? null;
    $mergeOffset = $rowParams["merge_offset"] ?? 0;
    if ($mergeCells !== null) {
      $currentRow = $this->currentRow;
      foreach ($mergeCells as [$leftCol, $rightCol]) {
        $this->mergeCells($leftCol + $mergeOffset, $currentRow, $rightCol + $mergeOffset, $currentRow);
      }
    }

    $this->ssWriter->addRow($row);
    $this->currentRow++;
    if ($differentOddEven) $this->oddEvenIndex++;
  }

  function writeHeaders(?array $headers=null): void {
    $this->rowStyle = self::STYLE_HEADER;
    parent::writeHeaders($headers);
    $this->rowStyle = self::STYLE_ROW;
  }

  function _sendContentType(): void {
    switch (path::ext($this->output)) {
    case ".ods":
      $contentType = "application/vnd.oasis.opendocument.spreadsheet";
      break;
    case ".xlsx":
    default:
      $contentType = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
      break;
    }
    http::content_type($contentType);
  }

  protected function _checkOk(): bool {
    $this->ssWriter->close();
    $this->rewind();
    return true;
  }
}
